import os

from qtpy.QtWidgets import QGridLayout, QWidget, QPushButton, QGroupBox, QHBoxLayout, QVBoxLayout
from tifffile import tifffile

from nellie import logger
from nellie.tracking.all_tracks_for_label import LabelTracks


class NellieVisualizer(QWidget):
    """
    The NellieVisualizer class provides an interface for visualizing different stages of the Nellie pipeline, such as raw images,
    preprocessed images, segmentation labels, mocap markers, and reassigned labels. It also enables track visualization for specific
    labels and all frame labels using napari.

    Attributes
    ----------
    nellie : object
        Reference to the Nellie instance managing the pipeline.
    viewer : napari.viewer.Viewer
        Reference to the napari viewer instance.
    scale : tuple of floats
        Scaling factors for visualizing the images, based on the resolution of the image dimensions.
    im_memmap : ndarray
        Memory-mapped array for the raw image.
    raw_layer : ImageLayer
        The napari image layer for displaying the raw image.
    im_branch_label_reassigned_layer : LabelsLayer
        The napari labels layer for displaying reassigned branch labels.
    im_branch_label_reassigned : ndarray
        Memory-mapped array for reassigned branch labels.
    im_obj_label_reassigned_layer : LabelsLayer
        The napari labels layer for displaying reassigned object labels.
    im_obj_label_reassigned : ndarray
        Memory-mapped array for reassigned object labels.
    im_marker_layer : ImageLayer
        The napari image layer for displaying mocap markers.
    im_marker : ndarray
        Memory-mapped array for mocap marker images.
    im_skel_relabelled_layer : LabelsLayer
        The napari labels layer for displaying skeleton relabeled images.
    im_instance_label_layer : LabelsLayer
        The napari labels layer for displaying instance labels.
    frangi_layer : ImageLayer
        The napari image layer for displaying preprocessed images (Frangi-filtered).
    im_skel_relabelled : ndarray
        Memory-mapped array for skeleton relabeled images.
    im_instance_label : ndarray
        Memory-mapped array for instance labels.
    im_frangi : ndarray
        Memory-mapped array for the preprocessed (Frangi-filtered) images.
    initialized : bool
        Flag indicating whether the visualizer has been initialized.

    Methods
    -------
    set_ui()
        Initializes and sets the layout and UI components for the NellieVisualizer.
    post_init()
        Initializes the visualizer by setting the scale and making the scale bar visible in napari.
    check_3d()
        Ensures that the napari viewer is in 3D mode if the dataset contains Z-dimension data.
    set_scale()
        Sets the scale for image display based on the image resolution in Z, Y, and X dimensions.
    open_preprocess_image()
        Opens and displays the preprocessed (Frangi-filtered) image in the napari viewer.
    open_segment_image()
        Opens and displays the segmentation labels (skeleton relabeled and instance labels) in the napari viewer.
    on_track_selected()
        Visualizes the tracks for the currently selected label in the napari viewer.
    track_all()
        Visualizes tracks for all labels across frames in the napari viewer.
    open_mocap_image()
        Opens and displays the mocap marker image in the napari viewer.
    open_reassign_image()
        Opens and displays the reassigned branch and object labels in the napari viewer.
    open_raw()
        Opens and displays the raw image in the napari viewer.
    check_file_existence()
        Checks for the existence of files related to different steps of the pipeline, enabling or disabling buttons accordingly.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        """
        Initializes the NellieVisualizer class, setting up buttons and layout for opening and visualizing images and tracks.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        nellie : object
            Reference to the Nellie instance managing the pipeline.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.scale = (1, 1, 1)

        self.im_memmap = None
        self.raw_layer = None
        self.im_branch_label_reassigned_layer = None
        self.im_branch_label_reassigned = None
        self.im_obj_label_reassigned_layer = None
        self.im_obj_label_reassigned = None
        self.im_marker_layer = None
        self.im_marker = None
        self.im_skel_relabelled_layer = None
        self.im_instance_label_layer = None
        self.frangi_layer = None
        self.im_skel_relabelled = None
        self.im_instance_label = None
        self.im_frangi = None

        # Raw im button
        self.raw_button = QPushButton(text="Open raw image")
        self.raw_button.clicked.connect(self.open_raw)
        self.raw_button.setEnabled(True)

        # Preprocess open button
        self.open_preprocess_button = QPushButton(text="Open preprocessed image")
        self.open_preprocess_button.clicked.connect(self.open_preprocess_image)
        self.open_preprocess_button.setEnabled(False)

        # Segment open button
        self.open_segment_button = QPushButton(text="Open segmentation images")
        self.open_segment_button.clicked.connect(self.open_segment_image)
        self.open_segment_button.setEnabled(False)

        # open im button
        self.open_mocap_button = QPushButton(text="Open mocap marker image")
        self.open_mocap_button.clicked.connect(self.open_mocap_image)
        self.open_mocap_button.setEnabled(False)

        # open reassign button
        self.open_reassign_button = QPushButton(text="Open reassigned labels image")
        self.open_reassign_button.clicked.connect(self.open_reassign_image)
        self.open_reassign_button.setEnabled(False)

        self.track_button = QPushButton(text="Visualize selected label's tracks")
        self.track_button.clicked.connect(self.on_track_selected)
        self.track_button.setEnabled(False)

        self.track_all_button = QPushButton(text="Visualize all frame labels' tracks")
        self.track_all_button.clicked.connect(self.track_all)
        self.track_all_button.setEnabled(False)

        self.set_ui()

        # self.layout = QGridLayout()
        # self.setLayout(self.layout)
        #
        # self.layout.addWidget(self.raw_button, 1, 0, 1, 1)
        # self.layout.addWidget(self.open_preprocess_button, 2, 0, 1, 1)
        # self.layout.addWidget(self.open_segment_button, 3, 0, 1, 1)
        # self.layout.addWidget(self.open_mocap_button, 4, 0, 1, 1)
        # self.layout.addWidget(self.open_reassign_button, 5, 0, 1, 1)

        self.initialized = False

    def set_ui(self):
        """
        Initializes and sets the layout and UI components for the NellieVisualizer. It groups the buttons for image
        and track visualization into separate sections and arranges them within a vertical layout.
        """
        main_layout = QVBoxLayout()

        # visualization group
        visualization_group = QGroupBox("Image visualization")
        visualization_layout = QVBoxLayout()
        visualization_layout.addWidget(self.raw_button)
        visualization_layout.addWidget(self.open_preprocess_button)
        visualization_layout.addWidget(self.open_segment_button)
        visualization_layout.addWidget(self.open_mocap_button)
        visualization_layout.addWidget(self.open_reassign_button)
        visualization_group.setLayout(visualization_layout)

        # tracking group
        tracking_group = QGroupBox("Track visualization")
        tracking_layout = QVBoxLayout()
        tracking_layout.addWidget(self.track_button)
        tracking_layout.addWidget(self.track_all_button)
        tracking_group.setLayout(tracking_layout)

        main_layout.addWidget(visualization_group)
        main_layout.addWidget(tracking_group)
        self.setLayout(main_layout)

    def post_init(self):
        """
        Post-initialization method that sets the image scale based on the image resolution and makes the scale bar visible.
        """
        self.set_scale()
        self.viewer.scale_bar.visible = True
        self.viewer.scale_bar.unit = 'um'
        self.initialized = True

    def check_3d(self):
        """
        Ensures that the napari viewer is in 3D mode if the dataset contains Z-dimension data.
        """
        if not self.nellie.im_info.no_z and self.viewer.dims.ndim != 3:
            # ndimensions should be 3 for viewer
            self.viewer.dims.ndim = 3
            self.viewer.dims.ndisplay = 3

    def set_scale(self):
        """
        Sets the scale for image display based on the resolution of the Z, Y, and X dimensions of the image.
        """
        dim_res = self.nellie.im_info.dim_res
        if self.nellie.im_info.no_z:
            self.scale = (dim_res['Y'], dim_res['X'])
        else:
            self.scale = (dim_res['Z'], dim_res['Y'], dim_res['X'])

    def open_preprocess_image(self):
        """
        Opens and displays the preprocessed (Frangi-filtered) image in the napari viewer.
        """
        self.im_frangi = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_preprocessed'])
        self.check_3d()
        self.frangi_layer = self.viewer.add_image(self.im_frangi, name='Pre-processed', colormap='turbo', scale=self.scale)
        self.frangi_layer.interpolation = 'nearest'
        self.viewer.layers.selection.active = self.frangi_layer

    def open_segment_image(self):
        """
        Opens and displays the segmentation labels (skeleton relabeled and instance labels) in the napari viewer.
        """
        self.im_instance_label = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_instance_label'])
        self.im_skel_relabelled = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_skel_relabelled'])

        self.check_3d()

        self.im_skel_relabelled_layer = self.viewer.add_labels(self.im_skel_relabelled, name='Labels: Branches',
                                                               opacity=1, scale=self.scale, visible=False)
        self.im_instance_label_layer = self.viewer.add_labels(self.im_instance_label, name='Labels: Organelles',
                                                              opacity=1, scale=self.scale)
        self.viewer.layers.selection.active = self.im_instance_label_layer

    def on_track_selected(self):
        """
        Visualizes the tracks for the currently selected label in the napari viewer, based on the active image layer.
        """
        # if flow_vector_array path does not point to an existing file, return
        if not os.path.exists(self.nellie.im_info.pipeline_paths['flow_vector_array']):
            return

        # pos = None
        # label = 0
        layer = self.viewer.layers.selection.active
        t = int(self.viewer.dims.current_step[0])
        # if event.button == 1 and event.is_dragging is False and 'Alt' in event.modifiers:
        #     scaled_pos = self.viewer.cursor.position
        #     pos = [scaled_pos[i+1] / self.scale[i] for i in range(len(scaled_pos)-1)]
        #     pos = (scaled_pos[0], *pos)
        #     pos = tuple(int(pos_dim) for pos_dim in pos)
        #     label = layer.selected_label
        # if pos is None:
        #     return


        if layer == self.im_instance_label_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        elif layer == self.im_skel_relabelled_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        elif layer == self.im_branch_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_branch_label_reassigned']
        elif layer == self.im_obj_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_obj_label_reassigned']
        else:
            return

        label = layer.selected_label
        if label == 0:
            return

        label_tracks = LabelTracks(im_info=self.nellie.im_info, label_im_path=label_path)
        label_tracks.initialize()
        all_tracks = []
        all_props = {}
        max_track_num = 0
        if self.nellie.settings.track_all_frames.isChecked() and (
                layer == self.im_branch_label_reassigned_layer or layer == self.im_obj_label_reassigned_layer):
            for frame in range(self.nellie.im_info.shape[0]):
                tracks, track_properties = label_tracks.run(label_num=label, start_frame=frame, end_frame=None,
                                                            min_track_num=max_track_num,
                                                            skip_coords=self.nellie.settings.skip_vox.value())
                all_tracks += tracks
                for property in track_properties.keys():
                    if property not in all_props.keys():
                        all_props[property] = []
                    all_props[property] += track_properties[property]
                if len(tracks) == 0:
                    break
                max_track_num = max([track[0] for track in tracks])+1
        else:
            all_tracks, all_props = label_tracks.run(label_num=label, start_frame=t, end_frame=None,
                                                     skip_coords=self.nellie.settings.skip_vox.value())
        if len(all_tracks) == 0:
            return
        self.viewer.add_tracks(all_tracks, properties=all_props, name=f'Tracks: Label {label}', scale=self.scale)
        self.viewer.layers.selection.active = layer
        self.check_file_existence()

    def track_all(self):
        """
        Visualizes tracks for all labels across frames in the napari viewer, based on the active image layer.
        """
        layer = self.viewer.layers.selection.active
        if layer == self.im_instance_label_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        elif layer == self.im_skel_relabelled_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        elif layer == self.im_branch_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_branch_label_reassigned']
        elif layer == self.im_obj_label_reassigned_layer:
            label_path = self.nellie.im_info.pipeline_paths['im_obj_label_reassigned']
        else:
            return

        t = int(self.viewer.dims.current_step[0])
        label_tracks = LabelTracks(im_info=self.nellie.im_info, label_im_path=label_path)
        label_tracks.initialize()
        all_tracks = []
        all_props = {}
        max_track_num = 0
        if self.nellie.settings.track_all_frames.isChecked() and (
                layer == self.im_branch_label_reassigned_layer or layer == self.im_obj_label_reassigned_layer):
            for frame in range(self.nellie.im_info.shape[0]):
                tracks, track_properties = label_tracks.run(start_frame=frame, end_frame=None,
                                                            min_track_num=max_track_num,
                                                            skip_coords=self.nellie.settings.skip_vox.value())
                all_tracks += tracks
                for property in track_properties.keys():
                    if property not in all_props.keys():
                        all_props[property] = []
                    all_props[property] += track_properties[property]
                if len(tracks) == 0:
                    break
                max_track_num = max([track[0] for track in tracks]) + 1
        else:
            all_tracks, all_props = label_tracks.run(start_frame=t, end_frame=None,
                                                     skip_coords=self.nellie.settings.skip_vox.value())
        if len(all_tracks) == 0:
            return
        self.viewer.add_tracks(all_tracks, properties=all_props, name=f'Tracks: All labels', scale=self.scale)
        self.viewer.layers.selection.active = layer
        self.check_file_existence()

    def open_mocap_image(self):
        """
        Opens and displays the mocap marker image in the napari viewer.
        """
        self.check_3d()

        self.im_marker = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_marker'])
        self.im_marker_layer = self.viewer.add_image(self.im_marker, name='Mocap Markers', colormap='red',
                                                     blending='additive', contrast_limits=[0, 1], scale=self.scale)
        self.im_marker_layer.interpolation = 'nearest'
        self.viewer.layers.selection.active = self.im_marker_layer

    def open_reassign_image(self):
        """
        Opens and displays the reassigned branch and object labels in the napari viewer.
        """
        self.check_3d()

        self.im_branch_label_reassigned = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_branch_label_reassigned'])
        self.im_branch_label_reassigned_layer = self.viewer.add_labels(self.im_branch_label_reassigned, name='Reassigned px: Branches', scale=self.scale, visible=False)

        self.im_obj_label_reassigned = tifffile.memmap(self.nellie.im_info.pipeline_paths['im_obj_label_reassigned'])
        self.im_obj_label_reassigned_layer = self.viewer.add_labels(self.im_obj_label_reassigned, name='Reassigned px: Organelles', scale=self.scale)
        self.viewer.layers.selection.active = self.im_obj_label_reassigned_layer

    def open_raw(self):
        """
        Opens and displays the raw image in the napari viewer.
        """
        self.check_3d()
        self.im_memmap = tifffile.memmap(self.nellie.im_info.im_path)
        self.raw_layer = self.viewer.add_image(self.im_memmap, name='raw', colormap='gray',
                                               blending='additive', scale=self.scale)
        # make 3d interpolation to nearest
        self.raw_layer.interpolation = 'nearest'
        self.viewer.layers.selection.active = self.raw_layer

    def check_file_existence(self):
        """
        Checks for the existence of files related to different steps of the pipeline, enabling or disabling buttons accordingly.
        """
        # set all other buttons to disabled first
        self.raw_button.setEnabled(False)
        self.open_preprocess_button.setEnabled(False)
        self.open_segment_button.setEnabled(False)
        self.open_mocap_button.setEnabled(False)
        self.open_reassign_button.setEnabled(False)
        self.track_button.setEnabled(False)
        self.track_all_button.setEnabled(False)

        if os.path.exists(self.nellie.im_info.im_path):
            self.raw_button.setEnabled(True)
        else:
            self.raw_button.setEnabled(False)

        frangi_path = self.nellie.im_info.pipeline_paths['im_preprocessed']
        if os.path.exists(frangi_path):
            self.open_preprocess_button.setEnabled(True)
        else:
            self.open_preprocess_button.setEnabled(False)

        im_instance_label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        im_skel_relabelled_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        if os.path.exists(im_instance_label_path) and os.path.exists(im_skel_relabelled_path):
            self.open_segment_button.setEnabled(True)
            self.nellie.settings.skip_vox.setEnabled(True)
            self.nellie.settings.track_all_frames.setEnabled(True)
            self.track_button.setEnabled(True)
            self.track_all_button.setEnabled(True)
            # self.viewer.help = 'Alt + click a label to see its tracks'
        else:
            self.open_segment_button.setEnabled(False)
            self.nellie.settings.skip_vox.setEnabled(False)
            self.nellie.settings.track_all_frames.setEnabled(False)

        im_marker_path = self.nellie.im_info.pipeline_paths['im_marker']
        if os.path.exists(im_marker_path):
            self.open_mocap_button.setEnabled(True)
        else:
            self.open_mocap_button.setEnabled(False)

        im_branch_label_path = self.nellie.im_info.pipeline_paths['im_branch_label_reassigned']
        if os.path.exists(im_branch_label_path):
            self.open_reassign_button.setEnabled(True)
            self.track_button.setEnabled(True)
            self.track_all_button.setEnabled(True)
            # self.viewer.help = 'Alt + click a label to see its tracks'
        else:
            self.open_reassign_button.setEnabled(False)


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import os
import subprocess
import time

from napari.utils.notifications import show_info
from qtpy.QtWidgets import QWidget, QPushButton, QVBoxLayout, QGroupBox, QLabel, QHBoxLayout
from qtpy.QtGui import QFont, QIcon
from qtpy.QtCore import Qt, QTimer

from nellie.feature_extraction.hierarchical import Hierarchy
from nellie.segmentation.filtering import Filter
from nellie.segmentation.labelling import Label
from nellie.segmentation.mocap_marking import Markers
from nellie.segmentation.networking import Network
from nellie.tracking.hu_tracking import HuMomentTracking
from nellie.tracking.voxel_reassignment import VoxelReassigner
from napari.qt.threading import thread_worker


class NellieProcessor(QWidget):
    """
    The NellieProcessor class manages the different steps of the Nellie pipeline such as preprocessing, segmentation,
    mocap marking, tracking, voxel reassignment, and feature extraction. It provides an interface to run each step
    individually or as part of a full pipeline within a napari viewer.

    Attributes
    ----------
    nellie : object
        Reference to the Nellie instance managing the pipeline.
    viewer : napari.viewer.Viewer
        Reference to the napari viewer instance.
    im_info_list : list of ImInfo or None
        List of ImInfo objects for the selected files. Contains metadata and file information.
    current_im_info : ImInfo or None
        The current image's information and metadata object.
    status_label : QLabel
        Label displaying the current status of the process.
    status : str or None
        The current status of the process (e.g., "preprocessing", "segmentation").
    num_ellipses : int
        Counter to manage the ellipsis effect on the status label during execution.
    status_timer : QTimer
        Timer that periodically updates the status label during pipeline execution.
    open_dir_button : QPushButton
        Button to open the output directory of the current image file.
    run_button : QPushButton
        Button to run the full Nellie pipeline.
    preprocess_button : QPushButton
        Button to run only the preprocessing step of the pipeline.
    segment_button : QPushButton
        Button to run only the segmentation step of the pipeline.
    mocap_button : QPushButton
        Button to run only the mocap marking step of the pipeline.
    track_button : QPushButton
        Button to run only the tracking step of the pipeline.
    reassign_button : QPushButton
        Button to run only the voxel reassignment step of the pipeline.
    feature_export_button : QPushButton
        Button to run only the feature extraction step of the pipeline.
    initialized : bool
        Flag indicating whether the processor has been initialized.
    pipeline : bool
        Flag indicating whether the full pipeline is being run.

    Methods
    -------
    set_ui()
        Initializes and sets the layout and UI components for the NellieProcessor.
    post_init()
        Post-initialization method to load image information and check file existence.
    check_file_existence()
        Checks the existence of necessary files for each step of the pipeline and enables/disables buttons accordingly.
    run_nellie()
        Runs the entire Nellie pipeline starting from preprocessing to feature extraction.
    run_preprocessing()
        Runs the preprocessing step of the Nellie pipeline.
    run_segmentation()
        Runs the segmentation step of the Nellie pipeline.
    run_mocap()
        Runs the mocap marking step of the Nellie pipeline.
    run_tracking()
        Runs the tracking step of the Nellie pipeline.
    run_reassign()
        Runs the voxel reassignment step of the Nellie pipeline.
    run_feature_export()
        Runs the feature extraction step of the Nellie pipeline.
    set_status()
        Sets the status to indicate that a process has started, and starts the status update timer.
    update_status()
        Updates the status label with the current process and an ellipsis effect while the process is running.
    reset_status()
        Resets the status label to indicate that no process is running.
    turn_off_buttons()
        Disables all buttons to prevent multiple processes from running simultaneously.
    open_directory()
        Opens the output directory where the current image results are saved.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        """
        Initializes the NellieProcessor class, setting up the user interface and preparing for running various steps of the pipeline.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        nellie : object
            Reference to the Nellie instance that manages the pipeline.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.im_info_list = None
        self.current_im_info = None

        self.status_label = QLabel("Awaiting your input")
        self.status = None
        self.num_ellipses = 1

        self.status_timer = QTimer()
        self.status_timer.timeout.connect(self.update_status)

        self.open_dir_button = QPushButton(text="Open output directory")
        self.open_dir_button.clicked.connect(self.open_directory)
        self.open_dir_button.setEnabled(False)

        # Run im button
        self.run_button = QPushButton(text="Run Nellie")
        self.run_button.clicked.connect(self.run_nellie)
        self.run_button.setEnabled(True)
        self.run_button.setFixedWidth(200)
        self.run_button.setFixedHeight(100)
        self.run_button.setStyleSheet("border-radius: 10px;")
        self.run_button.setFont(QFont("Arial", 20))

        # Preprocess im button
        self.preprocess_button = QPushButton(text="Run preprocessing")
        self.preprocess_button.clicked.connect(self.run_preprocessing)
        self.preprocess_button.setEnabled(False)

        # Segment im button
        self.segment_button = QPushButton(text="Run segmentation")
        self.segment_button.clicked.connect(self.run_segmentation)
        self.segment_button.setEnabled(False)

        # Run mocap button
        self.mocap_button = QPushButton(text="Run mocap marking")
        self.mocap_button.clicked.connect(self.run_mocap)
        self.mocap_button.setEnabled(False)

        # Run tracking button
        self.track_button = QPushButton(text="Run tracking")
        self.track_button.clicked.connect(self.run_tracking)
        self.track_button.setEnabled(False)

        # Run reassign button
        self.reassign_button = QPushButton(text="Run voxel reassignment")
        self.reassign_button.clicked.connect(self.run_reassign)
        self.reassign_button.setEnabled(False)

        # Run feature extraction button
        self.feature_export_button = QPushButton(text="Run feature export")
        self.feature_export_button.clicked.connect(self.run_feature_export)
        self.feature_export_button.setEnabled(False)

        self.set_ui()

        self.initialized = False
        self.pipeline = False

    def set_ui(self):
        """
        Initializes and sets the layout and user interface components for the NellieProcessor. This includes the status label,
        buttons for running individual steps, and the button for running the entire pipeline.
        """
        main_layout = QVBoxLayout()

        # Status group
        status_group = QGroupBox("Status")
        status_layout = QHBoxLayout()  # Changed to QHBoxLayout
        status_layout.addWidget(self.status_label, alignment=Qt.AlignLeft)
        status_layout.addWidget(self.open_dir_button, alignment=Qt.AlignCenter)
        status_group.setMaximumHeight(100)
        status_group.setLayout(status_layout)

        # Run full pipeline
        full_pipeline_group = QGroupBox("Run full pipeline")
        full_pipeline_layout = QVBoxLayout()
        full_pipeline_layout.addWidget(self.run_button, alignment=Qt.AlignCenter)
        full_pipeline_group.setLayout(full_pipeline_layout)

        # Run partial pipeline
        partial_pipeline_group = QGroupBox("Run individual steps")
        partial_pipeline_layout = QVBoxLayout()
        partial_pipeline_layout.addWidget(self.preprocess_button)
        partial_pipeline_layout.addWidget(self.segment_button)
        partial_pipeline_layout.addWidget(self.mocap_button)
        partial_pipeline_layout.addWidget(self.track_button)
        partial_pipeline_layout.addWidget(self.reassign_button)
        partial_pipeline_layout.addWidget(self.feature_export_button)
        partial_pipeline_group.setLayout(partial_pipeline_layout)

        main_layout.addWidget(status_group)
        main_layout.addWidget(full_pipeline_group)
        main_layout.addWidget(partial_pipeline_group)

        self.setLayout(main_layout)

    def post_init(self):
        """
        Post-initialization method that checks the state of the selected images. It determines whether the pipeline
        steps have already been completed and enables/disables the corresponding buttons accordingly.
        """
        if self.nellie.im_info_list is None:
            self.im_info_list = [self.nellie.im_info]
            self.current_im_info = self.nellie.im_info
        else:
            self.im_info_list = self.nellie.im_info_list
            self.current_im_info = self.nellie.im_info_list[0]
        self.check_file_existence()
        self.initialized = True
        self.open_dir_button.setEnabled(os.path.exists(self.current_im_info.file_info.output_dir))

    def check_file_existence(self):
        """
        Checks the existence of files required for each step of the pipeline (e.g., preprocessed images, segmented labels).
        Enables or disables buttons based on the existence of these files.
        """
        self.nellie.visualizer.check_file_existence()

        # set all other buttons to disabled first
        self.run_button.setEnabled(False)
        self.preprocess_button.setEnabled(False)
        self.segment_button.setEnabled(False)
        self.mocap_button.setEnabled(False)
        self.track_button.setEnabled(False)
        self.reassign_button.setEnabled(False)
        self.feature_export_button.setEnabled(False)

        analysis_path = self.current_im_info.pipeline_paths['features_organelles']
        if os.path.exists(analysis_path):
            self.nellie.setTabEnabled(self.nellie.analysis_tab, True)
        else:
            self.nellie.setTabEnabled(self.nellie.analysis_tab, False)

        self.open_dir_button.setEnabled(os.path.exists(self.current_im_info.file_info.output_dir))

        if os.path.exists(self.current_im_info.im_path):
            self.run_button.setEnabled(True)
            self.preprocess_button.setEnabled(True)
        else:
            return

        frangi_path = self.current_im_info.pipeline_paths['im_preprocessed']
        if os.path.exists(frangi_path):
            self.segment_button.setEnabled(True)
        else:
            self.segment_button.setEnabled(False)
            self.mocap_button.setEnabled(False)
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        im_instance_label_path = self.current_im_info.pipeline_paths['im_instance_label']
        im_skel_relabelled_path = self.current_im_info.pipeline_paths['im_skel_relabelled']
        if os.path.exists(im_instance_label_path) and os.path.exists(im_skel_relabelled_path):
            self.mocap_button.setEnabled(True)
        else:
            self.mocap_button.setEnabled(False)
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        im_marker_path = self.current_im_info.pipeline_paths['im_marker']
        if os.path.exists(im_marker_path):
            self.track_button.setEnabled(True)
        else:
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        track_path = self.current_im_info.pipeline_paths['flow_vector_array']
        if os.path.exists(track_path):
            self.reassign_button.setEnabled(True)
            self.feature_export_button.setEnabled(True)
        else:
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(True)
            # if im_info's 'T' axis has more than 1 timepoint, disable the feature export button
            if self.current_im_info.shape[0] > 1:
                self.feature_export_button.setEnabled(False)
                return



    @thread_worker
    def _run_preprocessing(self):
        """
        Runs the preprocessing step in a separate thread. Filters the image to remove noise or unwanted edges before segmentation.
        """
        self.status = "preprocessing"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Preprocessing file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            preprocessing = Filter(im_info=self.current_im_info,
                                   remove_edges=self.nellie.settings.remove_edges_checkbox.isChecked(),
                                   viewer=self.viewer)
            preprocessing.run()

    def run_preprocessing(self):
        """
        Starts the preprocessing step and updates the UI to reflect that preprocessing is running.
        If the full pipeline is running, it automatically proceeds to segmentation after preprocessing is finished.
        """
        worker = self._run_preprocessing()
        worker.started.connect(self.turn_off_buttons)
        if self.pipeline:
            worker.finished.connect(self.run_segmentation)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.finished.connect(self.check_file_existence)
        worker.start()

    @thread_worker
    def _run_segmentation(self):
        """
        Runs the segmentation step in a separate thread. Labels and segments regions of interest in the preprocessed image.
        """
        self.status = "segmentation"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Segmentation file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            segmenting = Label(im_info=self.current_im_info, viewer=self.viewer)
            segmenting.run()
            networking = Network(im_info=self.current_im_info, viewer=self.viewer)
            networking.run()

    def run_segmentation(self):
        """
        Starts the segmentation step and updates the UI to reflect that segmentation is running.
        If the full pipeline is running, it automatically proceeds to mocap marking after segmentation is finished.
        """
        worker = self._run_segmentation()
        worker.started.connect(self.turn_off_buttons)
        if self.pipeline:
            worker.finished.connect(self.run_mocap)
        worker.finished.connect(self.check_file_existence)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.start()

    @thread_worker
    def _run_mocap(self):
        """
        Runs the mocap marking step in a separate thread. Marks the motion capture points within the segmented regions.
        """
        self.status = "mocap marking"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Mocap Marking file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            mocap_marking = Markers(im_info=self.current_im_info, viewer=self.viewer)
            mocap_marking.run()

    def run_mocap(self):
        """
        Starts the mocap marking step and updates the UI to reflect that mocap marking is running.
        If the full pipeline is running, it automatically proceeds to tracking after mocap marking is finished.
        """
        worker = self._run_mocap()
        worker.started.connect(self.turn_off_buttons)
        if self.pipeline:
            worker.finished.connect(self.run_tracking)
        worker.finished.connect(self.check_file_existence)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.start()


    @thread_worker
    def _run_tracking(self):
        """
        Runs the tracking step in a separate thread. Tracks the motion of the marked points over time.
        """
        self.status = "tracking"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Tracking file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            hu_tracking = HuMomentTracking(im_info=self.current_im_info, viewer=self.viewer)
            hu_tracking.run()

    def run_tracking(self):
        """
        Starts the tracking step and updates the UI to reflect that tracking is running.
        If the full pipeline is running, it automatically proceeds to voxel reassignment or feature extraction depending on the settings.
        """
        worker = self._run_tracking()
        worker.started.connect(self.turn_off_buttons)
        if self.pipeline:
            if self.nellie.settings.voxel_reassign.isChecked():
                worker.finished.connect(self.run_reassign)
            else:
                worker.finished.connect(self.run_feature_export)
        worker.finished.connect(self.check_file_existence)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.start()

    @thread_worker
    def _run_reassign(self):
        """
        Runs the voxel reassignment step in a separate thread. Reassigns voxel labels based on the tracked motion.
        """
        self.status = "voxel reassignment"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Voxel Reassignment file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            vox_reassign = VoxelReassigner(im_info=self.current_im_info, viewer=self.viewer)
            vox_reassign.run()

    def run_reassign(self):
        """
        Starts the voxel reassignment step and updates the UI to reflect that voxel reassignment is running.
        If the full pipeline is running, it automatically proceeds to feature extraction after voxel reassignment is finished.
        """
        worker = self._run_reassign()
        worker.started.connect(self.turn_off_buttons)
        if self.pipeline:
            worker.finished.connect(self.run_feature_export)
        worker.finished.connect(self.check_file_existence)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.start()


    @thread_worker
    def _run_feature_export(self):
        """
        Runs the feature extraction step in a separate thread. Extracts various features from the processed image data for analysis.
        """
        self.status = "feature export"
        for im_num, im_info in enumerate(self.im_info_list):
            show_info(f"Nellie is running: Feature export file {im_num + 1}/{len(self.im_info_list)}")
            self.current_im_info = im_info
            hierarchy = Hierarchy(im_info=self.current_im_info,
                                  skip_nodes=not bool(self.nellie.settings.analyze_node_level.isChecked()),
                                  viewer=self.viewer)
            hierarchy.run()
            if self.nellie.settings.remove_intermediates_checkbox.isChecked():
                try:
                    self.current_im_info.remove_intermediates()
                except Exception as e:
                    show_info(f"Error removing intermediates: {e}")
        if self.nellie.analyzer.initialized:
            self.nellie.analyzer.rewrite_dropdown()

    def run_feature_export(self):
        """
        Starts the feature extraction step and updates the UI to reflect that feature extraction is running.
        """
        worker = self._run_feature_export()
        worker.started.connect(self.turn_off_buttons)
        worker.finished.connect(self.check_file_existence)
        worker.finished.connect(self.turn_off_pipeline)
        worker.started.connect(self.set_status)
        worker.finished.connect(self.reset_status)
        worker.start()

    def turn_off_pipeline(self):
        """
        Turns off the pipeline flag to indicate that the full pipeline is no longer running.
        """
        self.pipeline = False

    def run_nellie(self):
        """
        Starts the entire Nellie pipeline from preprocessing to feature extraction.
        """
        self.pipeline = True
        self.run_preprocessing()

    def set_status(self):
        """
        Sets the status of the processor to indicate that a process is running, and starts the status update timer.
        """
        self.running = True
        self.status_timer.start(500)  # Update every 250 ms

    def update_status(self):
        """
        Updates the status label with an ellipsis effect to indicate ongoing processing.
        """
        if self.running:
            self.status_label.setText(f"Running {self.status}{'.' * (self.num_ellipses % 4)}")
            self.num_ellipses += 1
        else:
            self.status_timer.stop()

    def reset_status(self):
        """
        Resets the status label to indicate that no process is running.
        """
        self.running = False
        self.status_label.setText("Awaiting your input")
        self.num_ellipses = 1
        self.status_timer.stop()

    def turn_off_buttons(self):
        """
        Disables all buttons to prevent multiple processes from running simultaneously.
        """
        self.run_button.setEnabled(False)
        self.preprocess_button.setEnabled(False)
        self.segment_button.setEnabled(False)
        self.mocap_button.setEnabled(False)
        self.track_button.setEnabled(False)
        self.reassign_button.setEnabled(False)
        self.feature_export_button.setEnabled(False)

    def open_directory(self):
        """
        Opens the output directory of the current image in the system file explorer.
        """
        directory = self.current_im_info.file_info.output_dir
        if os.path.exists(directory):
            if os.name == 'nt':  # For Windows
                os.startfile(directory)
            elif os.name == 'posix':  # For macOS and Linux
                subprocess.call(['open', directory])
        else:
            show_info("Output directory does not exist.")


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import os
import pickle

import numpy as np
from qtpy.QtWidgets import QComboBox, QCheckBox, QPushButton, QLabel, QTableWidget, QTableWidgetItem, \
    QSpinBox, QDoubleSpinBox, QWidget, QVBoxLayout, QGroupBox, QHBoxLayout
from qtpy.QtCore import Qt

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from napari.utils.notifications import show_info
import pandas as pd
import datetime


class NellieAnalysis(QWidget):
    """
    A class for analyzing and visualizing multi-dimensional microscopy data using histograms, graphs, and overlays in the napari viewer.

    Attributes
    ----------
    viewer : napari.viewer.Viewer
        An instance of the napari viewer.
    nellie : object
        Reference to the main Nellie object that contains the pipeline and analysis data.
    canvas : FigureCanvasQTAgg
        Canvas for rendering matplotlib plots.
    scale : tuple
        The scaling factors for X, Y, and Z dimensions, default is (1, 1, 1).
    log_scale : bool
        Boolean flag to toggle logarithmic scaling in histogram plots.
    is_median : bool
        Boolean flag to toggle between mean and median views.
    match_t : bool
        Boolean flag to enable/disable timepoint matching in data analysis.
    hist_reset : bool
        Boolean flag indicating whether the histogram settings are reset.
    voxel_df, node_df, branch_df, organelle_df, image_df : pd.DataFrame
        DataFrames containing features at different hierarchy levels (voxel, node, branch, organelle, image).
    attr_data : pd.Series or None
        Data for the selected attribute to be plotted.
    time_col : pd.Series or None
        Time column data from the currently selected level's DataFrame.
    adjacency_maps : dict or None
        Dictionary containing adjacency matrices for mapping hierarchy levels.
    mean, std, median, iqr, perc75, perc25 : float
        Statistical values for the selected attribute data (mean, std, median, interquartile range, 75th percentile, 25th percentile).

    Methods
    -------
    reset()
        Resets the internal state, including dataframes and initialization flags.
    post_init()
        Initializes UI elements such as checkboxes, buttons, and connects them to respective event handlers.
    set_ui()
        Sets up the user interface layout, including attribute selection, histogram, and export options.
    _create_dropdown_selection()
        Creates and configures the dropdowns for selecting hierarchy levels and attributes.
    set_default_dropdowns()
        Sets default values for the hierarchy level and attribute dropdowns.
    check_for_adjacency_map()
        Checks if an adjacency map is available and enables the overlay button accordingly.
    rewrite_dropdown()
        Rewrites the dropdown options based on the available data and features.
    export_data()
        Exports the current graph data to a CSV file.
    save_graph()
        Saves the current graph as a PNG file.
    on_hist_change(event)
        Event handler for histogram changes (e.g., adjusting bins or min/max values).
    get_index(layer, event)
        Gets the voxel index based on mouse hover coordinates in the napari viewer.
    overlay()
        Applies an overlay of selected attribute data onto the image, using adjacency maps to map voxel to higher-level features.
    on_t_change(event)
        Event handler that updates the graph when the timepoint changes in the napari viewer.
    toggle_match_t(state)
        Toggles timepoint matching and updates the graph accordingly.
    toggle_mean_med(state)
        Toggles between mean and median views and updates the graph.
    get_csvs()
        Loads the feature CSV files into DataFrames for voxels, nodes, branches, organelles, and images.
    on_level_selected(index)
        Event handler for when a hierarchy level is selected from the dropdown.
    on_attr_selected(index)
        Event handler for when an attribute is selected from the dropdown.
    get_stats()
        Computes basic statistics (mean, std, median, percentiles) for the selected attribute data.
    draw_stats()
        Draws the computed statistics on the histogram plot (e.g., mean, std, median, percentiles).
    plot_data(title)
        Plots the selected attribute data as a histogram, updates the UI, and displays statistical information.
    on_log_scale(state)
        Toggles logarithmic scaling for the histogram plot and refreshes the data.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        """
        Initializes the NellieAnalysis class.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        nellie : object
            Reference to the main Nellie object containing image and pipeline data.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.canvas = FigureCanvasQTAgg()
        self.canvas.figure.set_layout_engine("constrained")

        self.scale = (1, 1, 1)

        self.log_scale = False
        self.is_median = False
        self.mean_median_toggle = None
        self.overlay_button = None
        self.match_t_toggle = None
        self.match_t = False
        self.hist_min = None
        self.hist_max = None
        self.num_bins = None
        self.hist_reset = True
        self.export_data_button = None
        self.save_graph_button = None

        self.click_match_table = None

        self.voxel_df = None
        self.node_df = None
        self.branch_df = None
        self.organelle_df = None
        self.image_df = None

        self.all_attr_data = None
        self.attr_data = None
        self.time_col = None
        self.adjacency_maps = None
        self.data_to_plot = None
        self.mean = np.nan
        self.std = np.nan
        self.median = np.nan
        self.iqr = np.nan
        self.perc75 = np.nan
        self.perc25 = np.nan

        # self.layout = QGridLayout()
        # self.setLayout(self.layout)

        self.dropdown = None
        self.dropdown_attr = None
        self.log_scale_checkbox = None
        self.selected_level = None

        self.label_mask = None
        self.label_mask_layer = None
        self.label_coords = []

        self.click_match_texts = []

        self.layout_anchors = {
            'dropdown': (0, 0),
            'canvas': (1, 0),
            'table': (50, 0)
        }

        self.df = None
        self.initialized = False

    def reset(self):
        """
        Resets the internal state, including the DataFrames and initialization flags.
        """
        self.initialized = False
        self.voxel_df = None
        self.node_df = None
        self.branch_df = None
        self.organelle_df = None
        self.image_df = None

    def post_init(self):
        """
        Initializes UI elements such as checkboxes, buttons, and connects them to their respective event handlers.
        """
        self.log_scale_checkbox = QCheckBox("Log scale")
        self.log_scale_checkbox.stateChanged.connect(self.on_log_scale)

        self.mean_median_toggle = QCheckBox("Median view")
        self.mean_median_toggle.stateChanged.connect(self.toggle_mean_med)

        self.overlay_button = QPushButton("Overlay mask")
        self.overlay_button.clicked.connect(self.overlay)

        self.match_t_toggle = QCheckBox("Timepoint data")
        self.match_t_toggle.stateChanged.connect(self.toggle_match_t)
        self.match_t_toggle.setEnabled(False)
        self.viewer.dims.events.current_step.connect(self.on_t_change)

        self.hist_min = QDoubleSpinBox()
        self.hist_min.setEnabled(False)
        self.hist_min.valueChanged.connect(self.on_hist_change)
        self.hist_min.setDecimals(4)

        self.hist_max = QDoubleSpinBox()
        self.hist_max.setEnabled(False)
        self.hist_max.valueChanged.connect(self.on_hist_change)
        self.hist_max.setDecimals(4)

        self.num_bins = QSpinBox()
        self.num_bins.setRange(1, 100)
        self.num_bins.setValue(10)
        self.num_bins.setEnabled(False)
        self.num_bins.valueChanged.connect(self.on_hist_change)

        self.export_data_button = QPushButton("Export graph data")
        self.export_data_button.clicked.connect(self.export_data)

        self.save_graph_button = QPushButton("Save graph")
        self.save_graph_button.clicked.connect(self.save_graph)

        if self.nellie.im_info.no_z:
            self.scale = (self.nellie.im_info.dim_res['Y'], self.nellie.im_info.dim_res['X'])
        else:
            self.scale = (self.nellie.im_info.dim_res['Z'], self.nellie.im_info.dim_res['Y'], self.nellie.im_info.dim_res['X'])
        self.viewer.scale_bar.visible = True
        self.viewer.scale_bar.unit = 'um'

        self._create_dropdown_selection()
        self.check_for_adjacency_map()

        # # self.dropdown_attr = QComboBox()
        # self._create_dropdown_selection()
        # self.dropdown.setCurrentIndex(3)  # Organelle
        # self.dropdown_attr.currentIndexChanged.connect(self.on_attr_selected)
        # self.dropdown_attr.setCurrentIndex(6)  # Area

        self.set_ui()
        self.initialized = True

    def set_ui(self):
        """
        Sets up the user interface layout, including dropdowns for attribute selection, histogram controls, and export buttons.
        """
        main_layout = QVBoxLayout()

        # Attribute dropdown group
        attr_group = QGroupBox("Select hierarchy level and attribute")
        attr_layout = QHBoxLayout()
        attr_layout.addWidget(self.dropdown)
        attr_layout.addWidget(self.dropdown_attr)
        attr_group.setLayout(attr_layout)

        # Histogram group
        hist_group = QGroupBox("Histogram options")
        hist_layout = QVBoxLayout()

        sub_layout = QHBoxLayout()
        sub_layout.addWidget(QLabel("Min"), alignment=Qt.AlignRight)
        sub_layout.addWidget(self.hist_min)
        sub_layout.addWidget(self.hist_max)
        sub_layout.addWidget(QLabel("Max"), alignment=Qt.AlignLeft)
        hist_layout.addLayout(sub_layout)
        hist_layout.addWidget(self.canvas)

        sub_layout = QHBoxLayout()
        sub_layout.addWidget(QLabel("Bins"), alignment=Qt.AlignRight)
        sub_layout.addWidget(self.num_bins)
        hist_layout.addLayout(sub_layout)
        hist_group.setLayout(hist_layout)

        hist_layout.addWidget(self.canvas)

        sub_layout = QHBoxLayout()
        sub_layout.addWidget(self.log_scale_checkbox)
        sub_layout.addWidget(self.mean_median_toggle)
        sub_layout.addWidget(self.match_t_toggle)
        sub_layout.addWidget(self.overlay_button)
        hist_layout.addLayout(sub_layout)

        # Save options group
        save_group = QGroupBox("Export options")
        save_layout = QVBoxLayout()
        sub_layout = QHBoxLayout()
        sub_layout.addWidget(self.export_data_button)
        sub_layout.addWidget(self.save_graph_button)
        save_layout.addLayout(sub_layout)
        save_group.setLayout(save_layout)

        main_layout.addWidget(attr_group)
        main_layout.addWidget(hist_group)
        main_layout.addWidget(save_group)
        self.setLayout(main_layout)

    def _create_dropdown_selection(self):
        """
        Creates and configures the dropdown menus for selecting hierarchy levels and attributes.
        """
        # Create the dropdown menu
        self.dropdown = QComboBox()
        self.dropdown.currentIndexChanged.connect(self.on_level_selected)

        self.rewrite_dropdown()

        self.dropdown_attr = QComboBox()
        self.dropdown_attr.currentIndexChanged.connect(self.on_attr_selected)

        self.set_default_dropdowns()

    def set_default_dropdowns(self):
        """
        Sets the default values for the hierarchy level and attribute dropdowns.
        """
        organelle_idx = self.dropdown.findText('organelle')
        self.dropdown.setCurrentIndex(organelle_idx)
        area_raw_idx = self.dropdown_attr.findText('organelle_area_raw')
        self.dropdown_attr.setCurrentIndex(area_raw_idx)

    def check_for_adjacency_map(self):
        """
        Checks whether an adjacency map exists, and enables the overlay button if found.
        """
        self.overlay_button.setEnabled(False)
        if os.path.exists(self.nellie.im_info.pipeline_paths['adjacency_maps']):
            self.overlay_button.setEnabled(True)

    def rewrite_dropdown(self):
        """
        Updates the hierarchy level dropdown based on the available data, and checks for adjacency maps.
        """
        self.check_for_adjacency_map()

        self.dropdown.clear()
        if os.path.exists(self.nellie.im_info.pipeline_paths['features_nodes']):
            options = ['none', 'voxel', 'node', 'branch', 'organelle', 'image']
        else:
            options = ['none', 'voxel', 'branch', 'organelle', 'image']
        for option in options:
            self.dropdown.addItem(option)

        if self.dropdown_attr is not None:
            self.set_default_dropdowns()

        self.adjacency_maps = None

    def export_data(self):
        """
        Exports the current graph data as a CSV file to a specified directory.
        """
        dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        export_dir = self.nellie.im_info.graph_dir
        if not os.path.exists(export_dir):
            os.makedirs(export_dir)
        text = f"{dt}-{self.selected_level}-{self.dropdown_attr.currentText()}"
        if self.match_t:
            current_t = self.viewer.dims.current_step[0]
            text += f"_T{current_t}"
            timepoints = self.time_col[self.df['t'] == current_t]
        else:
            timepoints = self.time_col
        text += self.nellie.im_info.file_info.filename_no_ext
        export_path = os.path.join(export_dir, f"{text}.csv")
        df_to_save = pd.DataFrame({'t': timepoints, self.dropdown_attr.currentText(): self.attr_data})
        df_to_save.to_csv(export_path)

        # append time column
        show_info(f"Data exported to {export_path}")

    def save_graph(self):
        """
        Saves the current graph as a PNG file to a specified directory.
        """
        dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        export_dir = self.nellie.im_info.graph_dir
        if not os.path.exists(export_dir):
            os.makedirs(export_dir)
        text = f"{dt}-{self.selected_level}_{self.dropdown_attr.currentText()}"
        if self.match_t:
            text += f"_T{self.viewer.dims.current_step[0]}"
        text += self.nellie.im_info.file_info.filename_no_ext
        export_path = os.path.join(export_dir, f"{text}.png")
        self.canvas.figure.savefig(export_path, dpi=300)
        show_info(f"Graph saved to {export_path}")

    def on_hist_change(self, event):
        """
        Event handler for updating the histogram when changes are made (e.g., adjusting the number of bins, min/max values).
        """
        self.plot_data(self.dropdown_attr.currentText())

    def get_index(self, layer, event):
        """
        Retrieves the index of the voxel or feature based on mouse hover coordinates in the napari viewer.

        Parameters
        ----------
        layer : Layer
            The layer on which the event is triggered.
        event : Event
            The event triggered by mouse hover.
        """
        # get the coordinates of where the mouse is hovering
        pos = self.viewer.cursor.position
        matched_row = None
        if self.nellie.im_info.no_z:
            t, y, x = int(np.round(pos[0])), int(np.round(pos[1])), int(np.round(pos[2]))
        else:
            # todo, gonna be more complicated than this I think
            t, z, y, x = int(np.round(pos[0])), int(np.round(pos[1])), int(np.round(pos[2])), int(np.round(pos[3]))
        # find corresponding coord in self.label_coords
        if self.nellie.im_info.no_z:
            t_coords = self.label_coords[t]
            match = np.where((t_coords[:, 0] == y) & (t_coords[:, 1] == x))
            if len(match[0]) == 0:
                return
            else:
                matched_row = match[0][0]
            # show_info(f"Matched csv index: {matched_idx}, at T,Y,X: {t, y, x}")
        else:
            coord = self.label_coords[t][z]
        if matched_row is None:
            return
        # get the value of self.voxel_df_idxs[self.voxel_time_col == t] at the matched_idx row
        voxel_idx = self.voxel_df[self.voxel_df.columns[0]][self.voxel_df['t'] == t].iloc[matched_row]
        node_row = np.where(self.adjacency_maps['v_n'][t][matched_row])[0]
        node_idx = self.node_df[self.node_df.columns[0]][self.node_df['t'] == t].iloc[node_row].values
        branch_row = np.where(self.adjacency_maps['v_b'][t][matched_row])[0][0]
        branch_idx = self.branch_df[self.branch_df.columns[0]][self.branch_df['t'] == t].iloc[branch_row]
        organelle_row = np.where(self.adjacency_maps['v_o'][t][matched_row])[0][0]
        organelle_idx = self.organelle_df[self.organelle_df.columns[0]][self.organelle_df['t'] == t].iloc[organelle_row]
        image_row = np.where(self.adjacency_maps['v_i'][t][matched_row])[0][0]
        image_idx = self.image_df[self.image_df.columns[0]][self.image_df['t'] == t].iloc[image_row]

        self.click_match_table = QTableWidget()
        self.click_match_table.setRowCount(1)
        items = [f"{voxel_idx}", f"{node_idx}", f"{branch_idx}", f"{organelle_idx}", f"{image_idx}"]
        self.click_match_table.setColumnCount(len(items))
        if os.path.exists(self.nellie.im_info.pipeline_paths['features_nodes']):
            self.click_match_table.setHorizontalHeaderLabels(["Voxel", "Nodes", "Branch", "Organelle", "Image"])
        else:
            self.click_match_table.setHorizontalHeaderLabels(["Voxel", "Branch", "Organelle", "Image"])
        for i, item in enumerate(items):
            self.click_match_table.setItem(0, i, QTableWidgetItem(item))
        self.layout.addWidget(self.click_match_table, self.layout_anchors['table'][0], self.layout_anchors['table'][1], 1, 4)
        self.click_match_table.setVerticalHeaderLabels([f"{t, y, x}\nCSV row"])

    def overlay(self):
        """
        Applies an overlay of attribute data onto the image in the napari viewer, using adjacency maps for mapping between hierarchy levels.
        """
        if self.label_mask is None:
            label_mask = self.nellie.im_info.get_memmap(self.nellie.im_info.pipeline_paths['im_instance_label'])
            self.label_mask = (label_mask > 0).astype(float)
            for t in range(self.nellie.im_info.shape[0]):
                self.label_coords.append(np.argwhere(self.label_mask[t]))
                self.label_mask[t] *= np.nan

        if self.adjacency_maps is None:
            pkl_path = self.nellie.im_info.pipeline_paths['adjacency_maps']
            # load pkl file
            with open(pkl_path, 'rb') as f:
                adjacency_slices = pickle.load(f)
            self.adjacency_maps = {'n_v': [], 'b_v': [], 'o_v': []}
            for t in range(len(adjacency_slices['v_n'])):
                adjacency_slice = adjacency_slices['v_n'][t]
                # num_nodes = np.unique(adjacency_slice[:, 1]).shape[0]
                max_node = np.max(adjacency_slice[:, 1]) + 1
                min_node = np.min(adjacency_slice[:, 1])
                if len(self.label_coords[t]) == 0:
                    continue
                # adjacency_matrix = np.zeros((len(self.label_coords[t]), num_nodes), dtype=bool)
                adjacency_matrix = np.zeros((len(self.label_coords[t]), max_node), dtype=bool)
                adjacency_matrix[adjacency_slice[:, 0], adjacency_slice[:, 1]-min_node] = 1
                self.adjacency_maps['n_v'].append(adjacency_matrix.T)
            for t in range(len(adjacency_slices['v_b'])):
                adjacency_slice = adjacency_slices['v_b'][t]
                max_branch = np.max(adjacency_slice[:, 1]) + 1
                min_branch = np.min(adjacency_slice[:, 1])
                if len(self.label_coords[t]) == 0:
                    continue
                adjacency_matrix = np.zeros((len(self.label_coords[t]), max_branch), dtype=bool)
                adjacency_matrix[adjacency_slice[:, 0], adjacency_slice[:, 1]-min_branch] = 1
                self.adjacency_maps['b_v'].append(adjacency_matrix.T)
            for t in range(len(adjacency_slices['v_o'])):
                # organelles are indexed consecutively over the whole timelapse rather than per timepoint, so different
                #  way to construct the matrices
                adjacency_slice = adjacency_slices['v_o'][t]
                num_organelles = np.unique(adjacency_slice[:, 1]).shape[0]
                min_organelle = np.min(adjacency_slice[:, 1])
                if len(self.label_coords[t]) == 0:
                    continue
                # adjacency_matrix = np.zeros((np.max(adjacency_slice[:, 0])+1, np.max(adjacency_slice[:, 1])+1), dtype=bool)
                adjacency_matrix = np.zeros((len(self.label_coords[t]), num_organelles), dtype=bool)
                adjacency_matrix[adjacency_slice[:, 0], adjacency_slice[:, 1]-min_organelle] = 1
                self.adjacency_maps['o_v'].append(adjacency_matrix.T)
            # adjacency_slice = adjacency_slices['v_b'][t]
            # adjacency_matrix = np.zeros((adjacency_slice.shape[0], np.max(adjacency_slice[:, 1])+1))
            # adjacency_matrix[adjacency_slice[:, 0], adjacency_slice[:, 1]] = 1

        if self.attr_data is None:
            return

        for t in range(self.nellie.im_info.shape[0]):
            t_attr_data = self.all_attr_data[self.time_col == t].astype(float)
            if len(t_attr_data) == 0:
                continue
            if self.selected_level == 'voxel':
                self.label_mask[t][tuple(self.label_coords[t].T)] = t_attr_data
                continue
            elif self.selected_level == 'node' and len(self.adjacency_maps['n_v']) > 0:
                adjacency_mask = np.array(self.adjacency_maps['n_v'][t])
            elif self.selected_level == 'branch':
                adjacency_mask = np.array(self.adjacency_maps['b_v'][t])
            elif self.selected_level == 'organelle':
                adjacency_mask = np.array(self.adjacency_maps['o_v'][t])
            # elif self.selected_level == 'image':
            #     adjacency_mask = np.array(self.adjacency_maps['i_v'][t])
            else:
                return
            reshaped_t_attr = t_attr_data.values.reshape(-1, 1)
            # adjacency_mask = np.array(self.adjacency_maps['n_v'][t])
            if adjacency_mask.shape[0] == 0:
                continue
            attributed_voxels = adjacency_mask * reshaped_t_attr
            attributed_voxels[~adjacency_mask] = np.nan
            voxel_attributes = np.nanmean(attributed_voxels, axis=0)
            # if t > len(self.label_coords):
            #     continue
            # if self.selected_level == 'branch':
                # self.label_mask[t][tuple(self.skel_label_coords[t].T)] = voxel_attributes
            # else:
            self.label_mask[t][tuple(self.label_coords[t].T)] = voxel_attributes

        layer_name = f'{self.selected_level} {self.dropdown_attr.currentText()}'
        if 'reassigned' not in self.dropdown_attr.currentText():
            # label_mask_layer = self.viewer.add_image(self.label_mask, name=layer_name, opacity=1,
            #                                               colormap='turbo', scale=self.scale)
            perc98 = np.nanpercentile(self.attr_data, 98)
            # no nans, no infs
            real_vals = self.attr_data[~np.isnan(self.attr_data)]
            real_vals = real_vals[~np.isinf(real_vals)]
            min_val = np.min(real_vals) - (np.abs(np.min(real_vals)) * 0.01)
            if np.isnan(min_val):
                if len(real_vals) == 0:
                    min_val = 0
                else:
                    min_val = np.nanmin(real_vals)
            if min_val == perc98:
                perc98 = min_val + (np.abs(min_val) * 0.01)
            if np.isnan(perc98):
                if len(real_vals) == 0:
                    perc98 = 1
                else:
                    perc98 = np.nanmax(real_vals)
            contrast_limits = [min_val, perc98]
            # label_mask_layer.contrast_limits = contrast_limits
        # label_mask_layer.name = layer_name
        if not self.nellie.im_info.no_z:
            # if the layer isn't in 3D view, make it 3d view
            self.viewer.dims.ndisplay = 3
            # label_mask_layer.interpolation3d = 'nearest'
        # label_mask_layer.refresh()
        # self.label_mask_layer.mouse_drag_callbacks.append(self.get_index)
        if 'reassigned' in self.dropdown_attr.currentText():
            # make the label_mask_layer a label layer
                self.viewer.add_labels(self.label_mask.copy().astype('uint64'), scale=self.scale, name=layer_name)
        else:
            if not self.nellie.im_info.no_z:
                self.viewer.add_image(self.label_mask.copy(), name=layer_name, opacity=1,
                                      colormap='turbo', scale=self.scale, contrast_limits=contrast_limits,
                                      interpolation3d='nearest')
            else:
                self.viewer.add_image(self.label_mask.copy(), name=layer_name, opacity=1,
                                      colormap='turbo', scale=self.scale, contrast_limits=contrast_limits,
                                      interpolation2d='nearest')

        self.match_t_toggle.setEnabled(True)
        self.viewer.reset_view()

    def on_t_change(self, event):
        """
        Event handler for timepoint changes in the napari viewer. Updates the attribute data and refreshes the plot accordingly.
        """
        if self.match_t:
            self.on_attr_selected(self.dropdown_attr.currentIndex())

    def toggle_match_t(self, state):
        """
        Toggles timepoint matching and updates the graph and data accordingly.

        Parameters
        ----------
        state : int
            The state of the checkbox (checked or unchecked).
        """
        if state == 2:
            self.match_t = True
        else:
            self.match_t = False
        self.on_attr_selected(self.dropdown_attr.currentIndex())

    def toggle_mean_med(self, state):
        """
        Toggles between mean and median views for the histogram plot.

        Parameters
        ----------
        state : int
            The state of the checkbox (checked or unchecked).
        """
        if state == 2:
            self.is_median = True
        else:
            self.is_median = False
        self.on_attr_selected(self.dropdown_attr.currentIndex())

    def get_csvs(self):
        """
        Loads the CSV files containing voxel, node, branch, organelle, and image features into DataFrames.
        """
        self.voxel_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_voxels'])
        if os.path.exists(self.nellie.im_info.pipeline_paths['features_nodes']):
            self.node_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_nodes'])
        self.branch_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_branches'])
        self.organelle_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_organelles'])
        self.image_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_image'])

        # self.voxel_time_col = voxel_df['t']
        # self.voxel_df_idxs = voxel_df[voxel_df.columns[0]]

    def on_level_selected(self, index):
        """
        Event handler for when a hierarchy level is selected from the dropdown menu.

        Parameters
        ----------
        index : int
            The index of the selected item in the dropdown.
        """
        # This method is called whenever a radio button is selected
        # 'button' parameter is the clicked radio button
        self.selected_level = self.dropdown.itemText(index)
        self.overlay_button.setEnabled(True)
        if self.selected_level == 'voxel':
            if self.voxel_df is None:
                self.voxel_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_voxels'])
            self.df = self.voxel_df
        elif self.selected_level == 'node':
            if self.node_df is None:
                if os.path.exists(self.nellie.im_info.pipeline_paths['features_nodes']):
                    self.node_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_nodes'])
            self.df = self.node_df
        elif self.selected_level == 'branch':
            if self.branch_df is None:
                self.branch_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_branches'])
            self.df = self.branch_df
        elif self.selected_level == 'organelle':
            if self.organelle_df is None:
                self.organelle_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_organelles'])
            self.df = self.organelle_df
        elif self.selected_level == 'image':
            # turn off overlay button
            self.overlay_button.setEnabled(False)
            if self.image_df is None:
                self.image_df = pd.read_csv(self.nellie.im_info.pipeline_paths['features_image'])
            self.df = self.image_df
        else:
            return

        self.dropdown_attr.clear()
        # add a None option
        self.dropdown_attr.addItem("None")
        for col in self.df.columns[::-1]:
            # if "raw" not in col:
            #     continue
            # remove "_raw" from the column name
            # col = col[:-4]
            self.dropdown_attr.addItem(col)

    def on_attr_selected(self, index):
        """
        Event handler for when an attribute is selected from the dropdown menu.

        Parameters
        ----------
        index : int
            The index of the selected attribute in the dropdown.
        """
        self.hist_reset = True
        # if there are no items in dropdown_attr, return
        if self.dropdown_attr.count() == 0:
            return

        selected_attr = self.dropdown_attr.itemText(index)
        if selected_attr == '':
            return
        if selected_attr == "None":
            # clear the canvas
            self.canvas.figure.clear()
            self.canvas.draw()
            return

        self.all_attr_data = self.df[selected_attr]
        self.time_col = self.df['t']
        if self.match_t:
            t = self.viewer.dims.current_step[0]
            self.attr_data = self.df[self.df['t'] == t][selected_attr]
        else:
            self.attr_data = self.df[selected_attr]
        self.get_stats()
        self.plot_data(selected_attr)

    def get_stats(self):
        """
        Computes basic statistics (mean, std, median, percentiles) for the currently selected attribute data.
        """
        if self.attr_data is None:
            return
        if not self.log_scale:
            data = self.attr_data
        else:
            data = np.log10(self.attr_data)
            # convert non real numbers to nan
            data = data.replace([np.inf, -np.inf], np.nan)
        # only real data
        data = data.dropna()
        self.data_to_plot = data

        # todo only enable when mean is selected
        if not self.is_median:
            self.mean = np.nanmean(data)
            self.std = np.nanstd(data)

        # todo only enable when median is selected
        if self.is_median:
            self.median = np.nanmedian(data)
            self.perc75 = np.nanpercentile(data, 75)
            self.perc25 = np.nanpercentile(data, 25)
            self.iqr = self.perc75 - self.perc25

    def draw_stats(self):
        """
        Draws statistics on the histogram plot, including lines for mean, median, std, and percentiles.
        """
        if self.attr_data is None:
            return
        # draw lines for mean, median, std, percentiles on the canvas
        ax = self.canvas.figure.get_axes()[0]
        if self.is_median:
            ax.axvline(self.perc25, color='r', linestyle='--', label='25th percentile')
            ax.axvline(self.median, color='m', linestyle='-', label='Median')
            ax.axvline(self.perc75, color='r', linestyle='--', label='75th percentile')
        else:
            ax.axvline(self.mean - self.std, color='b', linestyle='--', label='Mean - Std')
            ax.axvline(self.mean, color='c', linestyle='-', label='Mean')
            ax.axvline(self.mean + self.std, color='b', linestyle='--', label='Mean + Std')
        ax.legend()
        self.canvas.draw()

    def plot_data(self, title):
        """
        Plots the selected attribute data as a histogram, updates the canvas, and displays the computed statistics.

        Parameters
        ----------
        title : str
            The title for the plot, usually the name of the selected attribute.
        """
        self.canvas.figure.clear()
        ax = self.canvas.figure.add_subplot(111)
        self.data_to_plot = self.data_to_plot.replace([np.inf, -np.inf], np.nan)
        try:
            if self.hist_reset:
                nbins = int(len(self.attr_data) ** 0.5)  # pretty nbins
                ax.hist(self.data_to_plot, bins=nbins)
                hist_min = ax.get_xlim()[0]
                hist_max = ax.get_xlim()[1]
            else:
                nbins = self.num_bins.value()
                hist_min = self.hist_min.value()
                hist_max = self.hist_max.value()
                ax.hist(self.data_to_plot, bins=nbins, range=(hist_min, hist_max))
        except ValueError:
            nbins = 10
            hist_min = 0
            hist_max = 1
        if self.is_median:
            full_title = f"{title}\n\nQuartiles: {self.perc25:.4f}, {self.median:.4f}, {self.perc75:.4f}"
        else:
            full_title = f"{title}\n\nMean: {self.mean:.4f}, Std: {self.std:.4f}"
        if self.match_t:
            full_title += f"\nTimepoint: {self.viewer.dims.current_step[0]}"
        else:
            full_title += f"\nTimepoint: all (pooled)"
        ax.set_title(full_title)
        if not self.log_scale:
            ax.set_xlabel("Value")
        else:
            ax.set_xlabel("Value (log10)")
        ax.set_ylabel("Frequency")
        self.canvas.draw()
        self.draw_stats()

        # if self.hist_min is not enabled
        if self.hist_reset:
            self.hist_min.setEnabled(True)
            self.hist_min.setValue(hist_min)
            self.hist_min.setRange(hist_min, hist_max)
            self.hist_min.setSingleStep((hist_max - hist_min) / 100)

            self.hist_max.setEnabled(True)
            self.hist_max.setValue(hist_max)
            self.hist_max.setRange(hist_min, hist_max)
            self.hist_max.setSingleStep((hist_max - hist_min) / 100)

            self.num_bins.setEnabled(True)
            self.num_bins.setValue(nbins)
            self.num_bins.setRange(1, len(self.attr_data))
            self.hist_reset = False

    def on_log_scale(self, state):
        """
        Toggles logarithmic scaling for the histogram plot and refreshes the data accordingly.

        Parameters
        ----------
        state : int
            The state of the checkbox (checked or unchecked).
        """
        self.hist_reset = True
        if state == 2:
            self.log_scale = True
        else:
            self.log_scale = False
        self.on_attr_selected(self.dropdown_attr.currentIndex())


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget, QPushButton, QMessageBox
from qtpy.QtGui import QPixmap, QFont
from qtpy.QtCore import Qt
import napari
import os
import datetime
import numpy as np
from napari.utils.notifications import show_info
import matplotlib.image


class Home(QWidget):
    """
    The Home screen for the Nellie application, displayed in the napari viewer.
    It provides options to start using the application, navigate to the file selection tab, and take screenshots.

    Attributes
    ----------
    viewer : napari.viewer.Viewer
        The napari viewer instance.
    nellie : object
        Reference to the main Nellie object containing image processing pipelines and functions.
    layout : QVBoxLayout
        The vertical layout to organize the widgets on the home screen.
    start_button : QPushButton
        Button to start the application and navigate to the file selection tab.
    screenshot_button : QPushButton
        Button to take a screenshot of the current napari viewer canvas.

    Methods
    -------
    __init__(napari_viewer, nellie, parent=None)
        Initializes the home screen with a logo, title, description, and navigation buttons.
    screenshot(event=None)
        Takes a screenshot of the napari viewer and saves it to a specified folder.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        """
        Initializes the Home screen with a logo, title, description, and buttons for navigation and screenshot functionality.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        nellie : object
            Reference to the main Nellie object containing image processing pipelines and functions.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        # Logo
        logo_path = os.path.join(os.path.dirname(__file__), 'logo.png')
        logo_label = QLabel(self)
        pixmap = QPixmap(logo_path)
        logo_label.setPixmap(
            pixmap.scaled(300, 300, Qt.KeepAspectRatio))#, Qt.SmoothTransformation))  # Adjust size as needed
        logo_label.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(logo_label)

        # Title
        title = QLabel("Nellie")
        title.setFont(QFont("Arial", 24, QFont.Bold))
        title.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(title)

        # Subtitle
        subtitle = QLabel("Automated organelle segmentation, tracking, and hierarchical feature extraction in 2D/3D live-cell microscopy.")
        subtitle.setFont(QFont("Arial", 16))
        subtitle.setAlignment(Qt.AlignCenter)
        subtitle.setWordWrap(True)
        self.layout.addWidget(subtitle)

        self.update_text = QLabel("Checking for updates...")
        self.update_text.setFont(QFont("Arial", 16))
        self.update_text.setStyleSheet("color: red")
        self.update_text.setAlignment(Qt.AlignCenter)
        self.update_text.setWordWrap(True)
        self.layout.addWidget(self.update_text)
        self.set_update_status()

        # Add a large "Start" button
        self.start_button = QPushButton("Start")
        self.start_button.setFont(QFont("Arial", 20))
        self.start_button.setFixedWidth(200)
        self.start_button.setFixedHeight(100)
        # rounded-edges
        self.start_button.setStyleSheet("border-radius: 10px;")
        # opens the file select tab
        self.start_button.clicked.connect(lambda: self.nellie.setCurrentIndex(self.nellie.file_select_tab))
        self.layout.addWidget(self.start_button, alignment=Qt.AlignCenter)

        github_link = QLabel("<a href='https://arxiv.org/abs/2403.13214'>Cite our paper!</a>")
        github_link.setOpenExternalLinks(True)
        github_link.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(github_link)

        # Link to GitHub
        github_link = QLabel("<a href='https://github.com/aelefebv/nellie'>Visit Nellie's GitHub Page!</a>")
        github_link.setOpenExternalLinks(True)
        github_link.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(github_link)

        # screenshot button
        self.screenshot_button = QPushButton(text="Easy screenshot:\n[Ctrl-Shift-E]")
        self.screenshot_button.setStyleSheet("border-radius: 5px;")
        self.screenshot_button.clicked.connect(self.screenshot)
        self.screenshot_button.setEnabled(True)
        self.viewer.bind_key('Ctrl-Shift-E', self.screenshot, overwrite=True)

        self.layout.addWidget(self.screenshot_button, alignment=Qt.AlignCenter)

    def set_update_status(self):
        """
        Checks if the plugin is up to date by comparing the installed version with the latest version on PyPI.
        If an update is available, it displays a warning to the user.
        """
        if self.nellie.current_version is None:
            self.update_text.setText("Failed to check for updates. \n")
        elif self.nellie.latest_version is None:
            self.update_text.setText("Failed to check for updates. \n"
                                     "Please check your internet connection.\n"
                                     f"Current: v{self.nellie.current_version}")
            return
        elif self.nellie.current_version == self.nellie.latest_version:
            self.update_text.setText("Nellie is up-to-date!\n"
                                     f"v{self.nellie.current_version}")
            # green
            self.update_text.setStyleSheet("color: green")
        else:
            self.update_text.setText(
                f"New version available!\n"
                f"Current: v{self.nellie.current_version}\n"
                f"Newest: v{self.nellie.latest_version}\n. "
                f"Please update to the latest version!"
            )
            self.update_text.setStyleSheet("color: red")


    def screenshot(self, event=None):
        """
        Takes a screenshot of the napari viewer and saves it as a PNG file in a specified folder.

        Parameters
        ----------
        event : optional
            An event object, if triggered by a key binding or button click (default is None).
        """
        if self.nellie.im_info is None:
            show_info("No file selected, cannot take screenshot")
            return

        # easy no prompt screenshot
        dt = datetime.datetime.now()  # year, month, day, hour, minute, second, millisecond up to 3 digits
        dt = dt.strftime("%Y%m%d_%H%M%S%f")[:-3]

        screenshot_folder = self.nellie.im_info.screenshot_dir
        if not os.path.exists(screenshot_folder):
            os.makedirs(screenshot_folder)

        im_name = f'{dt}-{self.nellie.im_info.file_info.filename_no_ext}.png'
        file_path = os.path.join(screenshot_folder, im_name)

        # Take screenshot
        screenshot = self.viewer.screenshot(canvas_only=True)

        # Save the screenshot
        try:
            # save as png to file_path using imsave
            screenshot = np.ascontiguousarray(screenshot)
            matplotlib.image.imsave(file_path, screenshot, format="png")
            show_info(f"Screenshot saved to {file_path}")
        except Exception as e:
            QMessageBox.warning(None, "Error", f"Failed to save screenshot: {str(e)}")
            raise e


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



from qtpy.QtWidgets import QWidget, QCheckBox, QSpinBox, QLabel, QVBoxLayout, QGroupBox, QHBoxLayout
import napari


class Settings(QWidget):
    """
    The Settings class provides a user interface for configuring various options and settings for the Nellie pipeline
    and visualizations. Users can enable or disable specific processing options, control track visualization settings,
    and configure voxel visualization parameters.

    Attributes
    ----------
    nellie : object
        Reference to the Nellie instance managing the pipeline.
    viewer : napari.viewer.Viewer
        Reference to the napari viewer instance.
    remove_edges_checkbox : QCheckBox
        Checkbox for enabling or disabling the removal of image edges.
    remove_intermediates_checkbox : QCheckBox
        Checkbox for enabling or disabling the removal of intermediate files after processing.
    voxel_reassign : QCheckBox
        Checkbox to enable or disable the automatic voxel reassignment step after tracking.
    analyze_node_level : QCheckBox
        Checkbox to enable or disable node-level analysis during feature extraction.
    track_all_frames : QCheckBox
        Checkbox to enable or disable the visualization of voxel tracks for all frames.
    skip_vox_label : QLabel
        Label describing the setting for skipping voxels during track visualization.
    skip_vox : QSpinBox
        Spin box for selecting the value of N to visualize tracks for every Nth voxel.
    initialized : bool
        Flag to indicate whether the settings interface has been initialized.

    Methods
    -------
    post_init()
        Post-initialization method that sets the initialized flag to True.
    set_ui()
        Initializes and sets the layout and UI components for the Settings class.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        """
        Initializes the Settings class, setting up the user interface and options for configuring
        processing and track visualization.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        nellie : object
            Reference to the Nellie instance that manages the pipeline.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        # Checkbox for 'Remove edges'
        self.remove_edges_checkbox = QCheckBox("Remove image edges")
        self.remove_edges_checkbox.setChecked(False)
        self.remove_edges_checkbox.setEnabled(True)
        self.remove_edges_checkbox.setToolTip(
            "Originally for Snouty deskewed images. If you see weird image edge artifacts, enable this.")

        self.remove_intermediates_checkbox = QCheckBox("Remove intermediate files")
        self.remove_intermediates_checkbox.setChecked(False)
        self.remove_intermediates_checkbox.setEnabled(True)
        self.remove_intermediates_checkbox.setToolTip(
            "Remove intermediate files after processing. This means only csv files will be saved.")

        self.voxel_reassign = QCheckBox("Auto-run voxel reassignment")
        self.voxel_reassign.setChecked(False)
        self.voxel_reassign.setEnabled(True)

        # Analyze node level
        self.analyze_node_level = QCheckBox("Analyze node level (slow)")
        self.analyze_node_level.setChecked(False)
        self.analyze_node_level.setEnabled(True)

        # Track all frames
        self.track_all_frames = QCheckBox("Visualize all frames' voxel tracks")
        self.track_all_frames.setChecked(True)
        self.track_all_frames.setEnabled(True)

        # Label above the spinner box
        self.skip_vox_label = QLabel("Visualize tracks for every N voxel. N=")

        self.skip_vox = QSpinBox()
        self.skip_vox.setRange(1, 10000)
        self.skip_vox.setValue(5)
        self.skip_vox.setEnabled(False)

        self.set_ui()

        self.initialized = False

    def post_init(self):
        """
        Post-initialization method that sets the initialized flag to True.
        """
        self.initialized = True

    def set_ui(self):
        """
        Initializes and sets the layout and UI components for the Settings class. This includes checkboxes for
        configuring the processing pipeline and track visualization options, as well as a spin box for setting
        voxel track visualization parameters.
        """
        main_layout = QVBoxLayout()

        # Processor settings
        processor_group = QGroupBox("Processor settings")
        processor_layout = QVBoxLayout()
        subprocessor_layout1 = QHBoxLayout()
        subprocessor_layout1.addWidget(self.remove_intermediates_checkbox)
        subprocessor_layout1.addWidget(self.remove_edges_checkbox)
        subprocessor_layout2 = QHBoxLayout()
        subprocessor_layout2.addWidget(self.analyze_node_level)
        subprocessor_layout2.addWidget(self.voxel_reassign)
        processor_layout.addLayout(subprocessor_layout1)
        processor_layout.addLayout(subprocessor_layout2)
        processor_group.setLayout(processor_layout)

        # Tracking settings
        tracking_group = QGroupBox("Track visualization settings")
        tracking_layout = QVBoxLayout()
        tracking_layout.addWidget(self.track_all_frames)
        skip_vox_layout = QHBoxLayout()
        skip_vox_layout.addWidget(self.skip_vox_label)
        skip_vox_layout.addWidget(self.skip_vox)
        tracking_layout.addLayout(skip_vox_layout)
        tracking_group.setLayout(tracking_layout)

        main_layout.addWidget(processor_group)
        main_layout.addWidget(tracking_group)
        self.setLayout(main_layout)


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



from napari.utils.notifications import show_info
from qtpy.QtWidgets import QTabWidget
import json
import urllib.request
from importlib.metadata import version as get_version
from napari.utils.notifications import show_warning

from nellie_napari import NellieProcessor
from nellie_napari.discover_plugins import add_nellie_plugins_to_menu
from nellie_napari.nellie_home import Home
from nellie_napari.nellie_analysis import NellieAnalysis
from nellie_napari.nellie_fileselect import NellieFileSelect
from nellie_napari.nellie_settings import Settings
from nellie_napari.nellie_visualizer import NellieVisualizer


class NellieLoader(QTabWidget):
    """
    The main loader class for managing the different stages of the Nellie pipeline within the napari viewer. This class
    provides a tabbed interface for file selection, processing, visualization, analysis, and settings management.

    Attributes
    ----------
    home : Home
        The home tab instance, providing an overview of the Nellie pipeline.
    file_select : NellieFileSelect
        The file selection tab instance, allowing users to select and validate image files.
    processor : NellieProcessor
        The image processing tab instance, where users can process images through the Nellie pipeline.
    visualizer : NellieVisualizer
        The visualization tab instance, where processed images can be visualized.
    analyzer : NellieAnalysis
        The analysis tab instance, enabling users to analyze processed image data.
    settings : Settings
        The settings tab instance, allowing users to configure various settings for the Nellie pipeline.
    home_tab, file_select_tab, processor_tab, visualizer_tab, analysis_tab, settings_tab : int
        Integer values representing the index of the respective tabs.
    im_info : ImInfo or None
        Contains metadata and information about the selected image file.
    im_info_list : list of ImInfo or None
        A list of ImInfo objects when batch processing is enabled (multiple files).

    Methods
    -------
    add_tabs()
        Adds the individual tabs to the widget.
    reset()
        Resets the state of the loader, removing and reinitializing all tabs.
    on_tab_change(index)
        Slot that is triggered when the user changes the tab.
    go_process()
        Initializes and enables the processing and visualization tabs for image processing.
    """
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', parent=None):
        """
        Initializes the NellieLoader class, creating instances of the individual tabs for home, file selection,
        processing, visualization, analysis, and settings.

        Parameters
        ----------
        napari_viewer : napari.viewer.Viewer
            Reference to the napari viewer instance.
        parent : QWidget, optional
            Optional parent widget (default is None).
        """
        super().__init__(parent)
        # Check for plugin updates
        self.current_version = None
        self.latest_version = None
        self.check_for_updates()

        self.viewer = napari_viewer
        self.home = Home(self.viewer, self)
        self.file_select = NellieFileSelect(self.viewer, self)
        self.processor = NellieProcessor(self.viewer, self)
        self.visualizer = NellieVisualizer(self.viewer, self)
        self.analyzer = NellieAnalysis(self.viewer, self)
        self.settings = Settings(self.viewer, self)

        self.home_tab = None
        self.file_select_tab = None
        self.processor_tab = None
        self.visualizer_tab = None
        self.analysis_tab = None
        self.settings_tab = None

        self.add_tabs()
        self.currentChanged.connect(self.on_tab_change)  # Connect the signal to the slot

        self.im_info = None
        self.im_info_list = None
        add_nellie_plugins_to_menu(self)

    def check_for_updates(self):
        """
        Gets the current installed version of the Nellie plugin and checks for the latest version available on PyPI.
        """
        show_info("Checking for updates...")
        try:
            # Get the current installed version
            self.current_version = get_version('nellie')

            # Fetch the latest version from PyPI
            with urllib.request.urlopen('https://pypi.org/pypi/nellie/json', timeout=5) as response:
                data = json.loads(response.read().decode())
                self.latest_version = data['info']['version']

        except Exception as e:
            pass

    def add_tabs(self):
        """
        Adds the individual tabs for Home, File validation, Process, Visualize, Analyze, and Settings.
        Initially disables the Process, Visualize, and Analyze tabs until they are needed.
        """
        ...
        self.home_tab = self.addTab(self.home, "Home")
        self.file_select_tab = self.addTab(self.file_select, "File validation")
        self.processor_tab = self.addTab(self.processor, "Process")
        self.visualizer_tab = self.addTab(self.visualizer, "Visualize")
        self.analysis_tab = self.addTab(self.analyzer, "Analyze")
        self.settings_tab = self.addTab(self.settings, "Settings")

        self.setTabEnabled(self.processor_tab, False)
        self.setTabEnabled(self.visualizer_tab, False)
        self.setTabEnabled(self.analysis_tab, False)

    def reset(self):
        """
        Resets the state of the loader, reinitializing all tabs. This method is typically called when the user
        wants to start a new session with a fresh file selection and settings.
        """
        self.setCurrentIndex(self.home_tab)

        # needs to be in reverse order
        self.removeTab(self.settings_tab)
        self.removeTab(self.analysis_tab)
        self.removeTab(self.visualizer_tab)
        self.removeTab(self.processor_tab)
        self.removeTab(self.file_select_tab)

        self.file_select = NellieFileSelect(self.file_select.viewer, self)
        self.processor = NellieProcessor(self.processor.viewer, self)
        self.visualizer = NellieVisualizer(self.visualizer.viewer, self)
        self.analyzer = NellieAnalysis(self.analyzer.viewer, self)
        self.settings = Settings(self.settings.viewer, self)

        self.add_tabs()

        self.im_info = None
        self.im_info_list = None

    def on_tab_change(self, index):
        """
        Event handler that is triggered when the user changes the active tab. Initializes the Analyze or Visualize
        tabs if they are selected for the first time, and always initializes the Settings tab.

        Parameters
        ----------
        index : int
            The index of the newly selected tab.
        """
        if index == self.analysis_tab:  # Check if the Analyze tab is selected
            if not self.analyzer.initialized:
                show_info("Initializing analysis tab")
                self.analyzer.post_init()
        elif index == self.visualizer_tab:
            if not self.visualizer.initialized:
                show_info("Initializing visualizer tab")
                self.visualizer.post_init()
        self.settings.post_init()

    def go_process(self):
        """
        Prepares the image(s) for processing and visualization. This method is called after a file has been selected
        and validated. It enables the Process and Visualize tabs and initializes them.
        """
        if self.file_select.batch_fileinfo_list is None:
            self.im_info = self.file_select.im_info
        else:
            self.im_info = self.file_select.im_info[0]
            self.im_info_list = self.file_select.im_info
            print(self.im_info_list)
        self.setTabEnabled(self.processor_tab, True)
        self.setTabEnabled(self.visualizer_tab, True)
        self.processor.post_init()
        self.visualizer.post_init()
        self.on_tab_change(self.processor_tab)
        self.setCurrentIndex(self.processor_tab)


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()
